/* -----------------------------------------

   Create deltaC x-tables for differential heavy-quark pair production

   ----------------------------------------- */
#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/time.h>
#include <getopt.h>

#include "hell.hh"
#include "QQbarPair.hh"


using namespace std;


string sas(double as) {
  ostringstream os;
  int ias = int(1000*as+0.1);
  if(1000*as-ias>0.5) ias += 1;
  if     (ias<10  ) os << "000" << ias;
  else if(ias<100 ) os << "00"  << ias;
  else if(ias<1000) os << "0"   << ias;
  else os << ias;
  return os.str();
}
string sx(double x) {
  ostringstream os;
  int ix = int(1e8*x+0.1);
  if(1e8*x-ix>0.5) ix += 1;
  if     (ix<1e1) os << "00000000"  << ix;
  else if(ix<1e2) os << "0000000"  << ix;
  else if(ix<1e3) os << "000000"  << ix;
  else if(ix<1e4) os << "00000"  << ix;
  else if(ix<1e5) os << "0000"  << ix;
  else if(ix<1e6) os << "000"  << ix;
  else if(ix<1e7) os << "00"  << ix;
  else if(ix<1e8) os << "0"   << ix;
  else os << ix;
  return os.str();
}


void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -x  --xx <x>           set x" << endl
       << " -a  --alphas <as>      set alpha_s" << endl
       << " -m  --invM <Q>         set the pair invariant mass" << endl
       << " -F  --muF <muF/Q>      set muF/Q" << endl
       << " -R  --RCmode           use the RC variant r->as*b0 (for estimating the uncertainty)" << endl
       << " -v  --variation        use the g+ variant \"Airy\" (for estimating the uncertainty)" << endl
       << " -r  --redo             do NOT recompute the resummed part with 2 M's" << endl
       << " -L  --LLp              use LLp (default is NLL)" << endl
       << endl;
  exit(0);
  return;
}

void read_arguments(int argc, char* argv[], double &x, double &as, double &mH, double &muFrat, int &mode, bool &redo, bool &useLLp, double &qt) {
  const char* const short_options = "hx:a:m:q:F:RvrL";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "xx", 1, NULL, 'x' },
					 { "alphas", 1, NULL, 'a' },
					 { "mH", 1, NULL, 'm' },
					 { "qt", 1, NULL, 'q' },
					 { "muF", 1, NULL, 'F' },
					 { "RCmode", 0, NULL, 'R' },
					 { "variation", 0, NULL, 'v' },
					 { "redo", 0, NULL, 'r' },
					 { "LLp", 0, NULL, 'L' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'x':
      x = strtod(optarg, NULL);
      break;
    case 'a':
      as = strtod(optarg, NULL);
      break;
    case 'm':
      mH = strtod(optarg, NULL);
      break;
    case 'q':
      qt = strtod(optarg, NULL);
      break;
    case 'F':
      muFrat = strtod(optarg, NULL);
      break;
    case 'v':
      mode = 1;
      break;
    case 'R':
      mode = 2;
      break;
    case 'r':
      redo = false;
      break;
    case 'L':
      useLLp = true;
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  return;
}



int main (int argc, char* argv[]) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  double m  = 4.6;
  double mH = 20;
  double x    = 1e-5;
  double rapY = 1;
  double pt   = 30;
  double as0 = 0.15;
  int nf = 4;
  int localRCmode = 0;
  double muFrat = 1;
  bool forceredo = true;
  bool useLLp = false;
  //
  read_arguments(argc, argv, x, as0, mH, muFrat, localRCmode, forceredo, useLLp, pt);

  if(x*(1+pt*pt/mH/mH) > 1) exit(555);
  double rapYmax = -log(x*(1+pt*pt/mH/mH))/2;
  //rapYmax *= 0.9999;
  double rapYmin = 0;


  HELL::SetLLpMode(useLLp);
  HELL::LogOrder ord = HELL::NLL;

  /*
  HELL::QQbarPair * pair = new HELL::QQbarPair(m);
  pair->SetKinematics(x, mH, rapY, pt);
  double xi1 = 2.2*pt*pt/mH/mH, xi2 = 2.2*pt*pt/mH/mH;
  double omega = 0.3;
  double beta = 0.8;
  double phi = 2.8;
  //double z1 = 0.3;
  //double z2 = 0.65;
  //double thetap = 1.3;
  //cout << setw(27) << pair->zzMel(xi1, 5, 2.5, z1, z2, thetap) << endl << endl;
  for(int i=8; i<20; i++) {
    //xi1 = 0.1*exp(-i*0.8);
    xi2 = 0.1*exp(-i*0.8);
    cout << setw(12) << xi2
	 << setw(15) << pair->TESTzzdCoff_dphi_int(xi1, xi2, omega, beta)
	 << setw(15) << pair->dCoff_dphi(xi1, xi2, phi, omega, beta) << endl;
  }
  cout << setw(27) << pair->Coff(xi1, omega, beta) << endl
       << setw(27) << pair->dCoff_dphi_0(xi1, omega, beta)*2*M_PI
       << setw(15) << pair->dCoff_dphi_0(xi1, phi, omega, beta) << endl << endl;
  //cout << setw(27) << pair->zzMel0(xi1, z1, z2, thetap) << endl << endl;
  //
  //for(int i=8; i<20; i++) {
  //  xi1 = 0.1*exp(-i*0.8);
  //  cout << setw(12) << xi1 << setw(15) << pair->TESTzzCoff_int(xi1, beta) << endl;
  //}
  //cout << setw(27) << pair->zzC(beta) << endl;
  for(int i=8; i<20; i++) {
    xi1 = 0.1*exp(-i*1.1);
    cout << setw(12) << xi1
	 << setw(15) << pair->TESTzzCoff_int(xi1, beta)
	 << setw(15) << pair->Coff(xi1, omega, beta)
	 << setw(15) << pair->dCoff_dphi_0(xi1, omega, beta)*2*M_PI
	 << endl;
  }
  cout << setw(27) << pair->Con(beta)
       << setw(15) << pair->Con(beta) << endl << endl;
  exit(0);
  //*/


  // x grid
  const int Np = 40;

  cout << "nf = " << nf << endl;
  HELL::HELLnf sxD(nf, ord, "./data/");
  //sxD.SetLargexDamping();
  sxD.SetMomentumConservation(); // has no effect
  sxD.SetEvaluationMethod(1);
  sxD.SetFactorizationScale(muFrat);
  if(localRCmode) sxD.SetRCmode(1);
  vector<double> as1;
  sxD.GetAvailableAlphas(as1);
  //


  // alphas grid
  //double as_min = as1[0];
  //double as_max = as1[as1.size()-1];
  //if(as2[0]>as_min) as_min = as2[0];
  //if(as2[as2.size()-1]<as_max) as_max = as2[as2.size()-1];
  //int Nas = (as_max-as_min)*1000 +1;
  //Nas = (as_max-as_min)*100 +1; // smaller set; comment out for a (10 times) larger set
  //vector<double> asvals;
  //asvals.push_back(as_min);
  //while(asvals[asvals.size()-1]<as_max) {
    //asvals.push_back(asvals[i]+0.001);
    //asvals.push_back(asvals[asvals.size()-1]+0.01);
  //}
  //
  //for(int a=0; a<Nas; a++) {
  //double as = as_min + a*(as_max-as_min)/(Nas-1.);
  for(unsigned int a=0; a<as1.size(); a++) {
    double as = as1[a];
    if(as0>0 && fabs(as-as0)>1e-6) continue;
    if(as0>0) cout << "as = " << as << endl;
    as = int(1000*as+0.1)/1000.;
    //cout << setprecision(10) << 1000*as << "  " << sas(as) << endl;
    ostringstream filename;
    filename << "./output_for_HELLx/NLL" << (useLLp ? "-LLp" : "") << "/xtableCQQbarPairTripleDiff";
    if(localRCmode) filename << "var" << localRCmode;
    filename << "_muF" << 1000*muFrat << "_NLL_nf" << nf << "_alphas" << sas(as) << "_pt" << pt << "_x" << sx(x) << ".table";
    //
    ofstream ofile(filename.str().c_str());
    ofile << VERSIONx << endl;
    ofile
      << setw(15) << Np
      << setw(15) << x
      << setw(15) << rapYmax
      << setw(15) << rapYmin
      << endl;
    //
    double dCreg=0, dCaux=0, dCauxOrdAs=0, dCregOrdAs2=0;
    dCaux      = sxD.deltaQQbarPairAux     (as, x, m, mH, -rapY, pt);
    dCauxOrdAs = sxD.deltaQQbarPairAuxOrdAs(as, x, m, mH, -rapY, pt);
    ofile
      << setw(15) << dCaux
      << endl
      << setw(15) << dCauxOrdAs
      << endl;
    //
    //double x_min = 0.0001, x_max = 0.99;
    //cout << "ciao" << endl;
    //for(int i=370; i<Np; i++) {
    for(int i=0; i<Np; i++) {
      rapY = rapYmin + i*(rapYmax-rapYmin)/(Np-1.);
      //pt = 0.1+i*80./(Np1+Np2-1.);
      //if(i<Np1) pt = 0.1+i*20./(Np1-1.); else pt = 20.1+(i-Np1+1)*60./(Np2-0.);
      //x = x_min * exp(i/(Np-1.)*log(x_max/x_min));
      //
      if(i<Np-1) {
	dCreg       = sxD.deltaQQbarPair(as, x, m, mH,  rapY, pt);
	dCregOrdAs2 = 0;//sxD.deltaQQbarPairOrdAs2(as, x, m, mH,  rapY, pt);
      }
      //
      ofile
	<< setw(15) << rapY
	<< setw(15) << dCreg
	<< setw(15) << dCregOrdAs2
	<< endl;
      cout << "-";
      cout.flush();
    }
    cout << ".";
    cout.flush();
    //
    ofile.close();
  }
  cout << endl;


  // Finish time
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << "Total time: " << t << "s" << endl;

  return 0;

}
