/* -----------------------------------------

   Create deltaC x-tables for all available values of alphas

   ----------------------------------------- */
#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/time.h> 
#include <getopt.h>

#include "hell.hh"
#include "Pqg.hh"
#include "DIS.hh"


using namespace std;


string sas(double as) {
  ostringstream os;
  int ias = int(1000*as+0.1);
  if(1000*as-ias>0.5) ias += 1;
  if     (ias<10  ) os << "000" << ias;
  else if(ias<100 ) os << "00"  << ias;
  else if(ias<1000) os << "0"   << ias;
  else os << ias;
  return os.str();
}


void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -a  --alphas <as>      set alpha_s" << endl
       << " -n  --nf <nf>          set the number of active flavours" << endl
       << " -R  --RCmode           use the RC variant r->as*b0 (for estimating the uncertainty)" << endl
       << " -v  --variation        use the g+ variant \"Airy\" (for estimating the uncertainty)" << endl
       << " -L  --LLp              use LLp (default is NLL)" << endl
       << endl;
  exit(0);
  return;
}

void read_arguments(int argc, char* argv[], double &as, int &nf, int &mode, bool &useLLp) {
  const char* const short_options = "ha:n:RvL";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "alphas", 1, NULL, 'a' },
					 { "nf", 1, NULL, 'n' },
					 { "RCmode", 0, NULL, 'R' },
					 { "variation", 0, NULL, 'v' },
					 { "LLp", 0, NULL, 'L' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'a':
      as = strtod(optarg, NULL);
      break;
    case 'n':
      nf = int(strtod(optarg, NULL));
      break;
    case 'v':
      mode = 1;
      break;
    case 'R':
      mode = 2;
      break;
    case 'L':
      useLLp = true;
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  return;
}



int main (int argc, char* argv[]) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  bool useLLp = false;
  int nf0 = -1;
  double as0 = -1;
  int localRCmode = 0;
  read_arguments(argc, argv, as0, nf0, localRCmode, useLLp);

  HELL::SetLLpMode(useLLp);
  HELL::LogOrder ord = HELL::NLL;


  // x grid
  const int Np1 = 300, Np2 = 20-1, Nxi = 70;
  double xi[2*Nxi];
  double x[Np1+Np2];
  double x_min = 1e-9;
  double x_mid = 0.1;
  double x_max = 0.99+0.01;   /////////////////
  //
  for(int i=0; i<Np1+Np2; i++) {
    if(i<Np1) x[i] = x_min * exp(i/(Np1-1.)*log(x_mid/x_min));
    else      x[i] = x_mid + (i-Np1+1)*(x_max-x_mid)/(Np2-0.);
  }
  //
  for(int nf=3; nf<=6; nf++) {
    if(nf0>0 && nf!=nf0) continue;
    cout << "nf = " << nf << endl;
    HELL::HELLnf sxD(nf, ord, "./data/");
    //sxD.SetLargexDamping();
    //sxD.SetMomentumConservation(); // has no effect
    if(localRCmode==1) sxD.SetRCvariation();
    if(localRCmode==2) sxD.SetRCmode(1);
    vector<double> as1;
    sxD.GetAvailableAlphas(as1);
    // alphas grid
    //double as_min = as1[0];
    //double as_max = as1[as1.size()-1];
    //if(as2[0]>as_min) as_min = as2[0];
    //if(as2[as2.size()-1]<as_max) as_max = as2[as2.size()-1];
    //int Nas = (as_max-as_min)*1000 +1;
    //Nas = (as_max-as_min)*100 +1; // smaller set; comment out for a (10 times) larger set
    //vector<double> asvals;
    //asvals.push_back(as_min);
    //while(asvals[asvals.size()-1]<as_max) {
      //asvals.push_back(asvals[i]+0.001);
      //asvals.push_back(asvals[asvals.size()-1]+0.01);
    //}
    //
    //for(int a=0; a<Nas; a++) {
    //double as = as_min + a*(as_max-as_min)/(Nas-1.);
    //for(int a=0; a<asvals.size(); a++) {
    //  double as = asvals[a];
    for(unsigned int a=0; a<as1.size(); a++) {
      double as = as1[a];
      if(as0>0 && fabs(as-as0)>1e-6) continue;
      if(as0>0) cout << "as = " << as << endl;
      as = int(1000*as+0.1)/1000.;
      //cout << setprecision(10) << 1000*as << "  " << sas(as) << endl;
      //
      double logxi0 = -1./as/HELL::beta0(nf);
      int actualNxi = 0;
      /*
      for(int k=1; k<2*Nxi-3; k++) {
	if(k<Nxi) xi[k] = pow(k/(Nxi-1.),-3.);
	else      xi[k] = pow((2*Nxi-k-2)/(Nxi-1.),3.);
	if(xi[k]>=xi0) actualNxi++;
	else break;
      }
      */
      double logximax = 2 - 3*logxi0;
      if(logximax>70) logximax = 70;  // corresponds to ximax ~ 10^30
      for(int k=0; k<Nxi; k++) {
	xi[k] = exp(logxi0 + k/(Nxi-1.)*(-logxi0));
	actualNxi++;
      }
      for(int k=0; k<Nxi; k++) {
	xi[Nxi+k] = exp((k+1.)/Nxi*logximax);
	actualNxi++;
      }
      //
      ostringstream filename;
      filename << "./data/x-space-aux" << (useLLp ? "-LLp" : "") << "/xtableU";
      if(localRCmode) filename << "var" << localRCmode;
      filename << "_NLL_nf" << nf << "_alphas" << sas(as) << ".table";
      ofstream ofile(filename.str().c_str());
      ofile << VERSION << endl;
      ofile
	<< setw(15) << Np1
	<< setw(15) << Np2
	<< setw(15) << x_min
	<< setw(15) << x_mid
	<< setw(15) << x_max
	<< setw(15) << actualNxi
	<< endl;
      //
      for(int i=0; i<Np1+Np2; i++) {
	double dUqg = x[i]*sxD.deltaUqg(as, x[i]);
	ofile
	  << setw(15) << dUqg;
      }
      ofile << endl;
      cout << "-";
      cout.flush();
      for(int k=0; k<actualNxi; k++) {
	ofile << setw(15) << xi[k];
	for(int i=0; i<Np1+Np2; i++) {
	  double dU = x[i]*sxD.deltaU(as, x[i], xi[k]);
	  ofile
	    //<< setw(15) << x
	    << setw(15) << dU;
	}
	ofile << endl;
	cout << ".";
	cout.flush();
      }
      /*
      // this is not better than computing the derivative numerically, because the U' oscillates more than U, so the interpolation is better for U than U'
      // derivative
      for(int k=0; k<actualNxi; k++) {
	ofile << setw(15) << xi[k];
	for(int i=0; i<Np1+Np2; i++) {
	  double ddU = x[i]*sxD.deltaDU(as, x[i], xi[k]);
	  ofile
	    //<< setw(15) << x
	    << setw(15) << ddU;
	}
	ofile << endl;
	cout << ".";
	cout.flush();
      }
      cout << ".";
      cout.flush();
      */
      //
      ofile.close();
    }
    cout << endl;
  }


  // Finish time
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << "Total time: " << t << "s" << endl;

  return 0;

}
