#include <hell_tables.hh>
#include <integration.hh>
#include <gsl/gsl_errno.h>
#include <cstdlib>
#include "version.hh"

namespace HELL {
  
  //const Real HELL_N_INTERVALS = 10000;
  //const Real HELL_N_STEP = 0.0000998;
  
  GammaHellTable & GammaHellTable::read(std::istream & in)
  {
    Real tmp;
    in >> version_ >> alphas_ >> nf_;
    if(version_!=VERSION) {
      std::cout << "\033[0;31m" << "HELL: Error! The tables you are trying to read are version " << version_ << " which is not compatible with the code version " << VERSION << "\033[0m" << std::endl;
      abort();
    }
    //
    int npoles;
    in >> npoles;
    poles_.resize(npoles);
    for (int i=0; i<npoles; ++i)
      in >> poles_[i].pole;
    for (int i=0; i<npoles; ++i)
      in >> poles_[i].residue;
    //
    int npoints;
    in >> npoints >> umax_;
    step_ = umax_/(npoints-1.);
    gamma_   .resize(npoints);
    gammavar_.resize(npoints);
    in >> N0_;
    in >> tmp;  eith_ = exp(I*PI*tmp);
    //
    for (int i=0; i<npoints; ++i) {
      in >> tmp;  gamma_[i]  =   tmp;
      in >> tmp;  gamma_[i] += I*tmp;
      in >> tmp;  gammavar_[i]  =   tmp;
      in >> tmp;  gammavar_[i] += I*tmp;
    }
    return *this;
  }
  Complex GammaHellTable::getN(unsigned i) const {
    Real u = i*step_;
    Real t = u/(1-u);
    return N0_ + t*eith_;
  }



  /// HellTableP

  HellTableP & HellTableP::readTable(std::string in) {
    std::ifstream infile(in.c_str());
    if (!infile.is_open()) {
      std::cerr << "\033[0;31m" << "HELL: Error reading table " << in << "\033[0m" << std::endl;
      exit(-4);
    }
    readTable(infile);
    return *this;
  }
  HellTableP & HellTableP::readDotTable(std::string in) {
    std::ifstream infile(in.c_str());
    if (!infile.is_open())
      std::cerr << "\033[0;31m" << "HELL: Error reading table " << in << "\033[0m" << std::endl;
    readDotTable(infile);
    return *this;
  }
  //HellTableP & HellTableP::readNLOTable(std::string in) {
  //  std::ifstream infile(in.c_str());
  //  if (!infile.is_open())
  //    std::cerr << "\033[0;31m" << "HELL: Error reading table " << in << "\033[0m" << std::endl;
  //  readNLOTable(infile);
  //  return *this;
  //}

  unsigned HellTableP::getTableSize() const {
    unsigned size = g_.getTableSize();
    if(g_dot_.getTableSize()>0) size = std::min(size,g_dot_.getTableSize());
    //if(g_nlo_.getTableSize()>0) size = std::min(size,g_nlo_.getTableSize());
    return size;
  }

  HellTableP & HellTableP::setMellinFunction(MellinFunction mf, void *pars) {
    int n = getTableSize();
    MFvals_.resize(n+1);
    double *u    = new double[n+2];
    double *reMF = new double[n+2];
    double *imMF = new double[n+2];
    //
    for (int i=0; i<n+1; ++i) {
      //for (int i=1170; i<1171; ++i) {
      u[i] = i*g_.step_;
      Complex N = getN(i);
      Complex g     = getGamma(i);
      Complex g_dot = getDotGamma(i);
      //Complex g_nlo = getNLOGamma(i);
      //std::cout << i << std::endl;
      //MFvals_[i] = (*mf)(g, g_dot, N, getAlphaS(), getNF(), g_nlo, pars);
      MFvals_[i] = (*mf)(g, g_dot, N, getAlphaS(), getNF(), pars);
      reMF[i] = real(MFvals_[i]);
      imMF[i] = imag(MFvals_[i]);
    }
    u   [n+1] = 1;
    reMF[n+1] = 0;
    imMF[n+1] = 0;
    //
    // spline setup
    reMFspline_ = gsl_spline_alloc (gsl_interp_cspline, n+2);
    imMFspline_ = gsl_spline_alloc (gsl_interp_cspline, n+2);
    gsl_spline_init (reMFspline_, u, reMF, n+2);
    gsl_spline_init (imMFspline_, u, imMF, n+2);
    //
    delete[] u;
    delete[] reMF;
    delete[] imMF;
    //
    return *this;
  }
  Complex HellTableP::getMellinFunctionSpline(double u) {
    gsl_interp_accel *acc = gsl_interp_accel_alloc ();
    double reMF = gsl_spline_eval(reMFspline_, u , acc);
    double imMF = gsl_spline_eval(imMFspline_, u , acc);
    delete acc;
    return Complex(reMF,imMF);
  }


  Real HellTableP::evalSimpson(Real z)
  {
    Real z_max = inv_mellin_max_z(z);
    Real u_max = z_max/(1+z_max);
    const Real TABLE_U_MAX = g_.step_*(getTableSize()-1);
    if(u_max > TABLE_U_MAX) u_max = TABLE_U_MAX;
    Real n = 4*static_cast<int>(u_max/g_.step_*0.25) + 4;
    //Real n = 6*static_cast<int>(z_max/g_.step_/6) + 6;
    u_max = n*g_.step_;
    if(u_max > TABLE_U_MAX) n -= 4;
    u_max = n*g_.step_;
    if(u_max > TABLE_U_MAX) abort(); // it should never happen
    //
    std::vector<Real> fvals(n+1);
    //Real *fvals = new Real[(int)n+1];
    for (int i=0; i<n+1; ++i) {
      fvals[i] = imag(g_.eith_*pow(z,-getN(i)) * MFvals_[i]) * getJac(i);
      /*
      double res[3], epsilon = 1.e-4;
      for(int k=0; k<3; k++) {
	res[k] = imag(g_.eith_*pow(z,-getN(i)) * MFvals_[i]) * getJac(i) * exp(-epsilon*pow(imag(getN(i)),2));
	epsilon /= 2.;
      }
      fvals[i] = (res[0] - 6*res[1] + 8*res[2])/3.;
      */
    }
    Real * err_ptr = use_err_ ? &err_ : 0;
    //
    return SimpsonIntegration::integrate(&fvals[0],0.0,u_max,n,err_ptr)/PI;
  }

  Real HellTableP::GaussIntegrand(Real t) {
    Complex nt = g_.N0_ + t*g_.eith_;
    Real u = t/(1+t);
    /*
    const Real TABLE_U_MAX = g_.step_*(getTableSize()-1);
    if(u>TABLE_U_MAX) return 0;
    double ii = u/g_.step_;
    int i = (int)ii; // i <= ii
    //std::cout << i << "  " << ii << std::endl;
    //Real u = i*g_.step_;
    Real factor = ii-i;
    Complex gg = (1-factor)*MFvals_[i] + factor*MFvals_[i+1]; // interpolation
    */
    Complex gg = getMellinFunctionSpline(u);
    //double epsilon = 1.e-2;
    Real res = imag(g_.eith_*pow(x_,-nt-1.) * gg);// * exp(-epsilon*pow(imag(getN(i)),2));
    //std::cout << res << std::endl;
    return res;
  }
  double integrand(double x, void *p) {
    HellTableP *htp = (HellTableP *) p;
    //double t = x, jac = 1;
    //double t = x/(1-x), jac = 1/(1-x)/(1-x);
    //double t = -log(1-x), jac = 1/(1-x);
    double t = -log(1-x*x), jac = 2*x/(1-x*x);
    return htp->GaussIntegrand(t) * jac;
  }
  Real HellTableP::evalGauss(Real z, int &status) {
    x_ = z;
    double prec = 5e-3, error;
    /*
    double N=1000;
    std::ofstream ofile("output/test_integrand_Mellin_inversion.dat");
    ofile << z << std::endl;
    for(int i=0; i<N; i++) {
      double x = i/(N-1.);
      ofile << std::setw(12) << x << std::setw(16) << z*integrand(x,this) << std::endl;
    }
    std::cout << z << "  " << z*integration_gauss(integrand, 0., 1, prec, &error, this, &status) << std::endl;
    if(z>0.00001) exit(1);
    //*/
    //return z*integration_gauss(integrand, 0., 33./log(1./z), prec, &error, this, &status)/PI;
    return z*integration_gauss(integrand, 0., 1, prec, &error, this, &status)/PI;
  }


  Real HellTableP::getIntegrand(Real z, int i) {
    Real dt = getJac(i);
    Complex nt = getN(i);
    return imag(g_.eith_*pow(z,-nt) * MFvals_[i])*dt;
  }
  Real HellTableP::getIntegrand(MellinFunction mf, Real z, int i, void *pars) {
    Real dt = getJac(i);
    Complex nt = getN(i);
    Complex g = getGamma(i);
    Complex g_dot = getDotGamma(i);
    if (mf)
      g = (*mf)(g, g_dot, nt, getAlphaS(), getNF(), pars);
    //
    return imag(g_.eith_*pow(z,-nt) * g)*dt;
  }

  Real HellTableP::eval(Real z) {
    //return evalSimpson(z);
    //if(z<5e-5) return evalSimpson(z);
    Real result = 0;
    // I'd like to compute the integral with gauss (faster) and, if there is an error, ignore it and use Simpson (slower).
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    int status;
    result = evalGauss(z, status);
    gsl_set_error_handler (old_handler);
    if(status) {
      //std::cout << z << std::endl;
      //result = evalSimpson(z);
    }
    return result;
  }

  Real HellTableP::getJac(unsigned i) const {
    Real u = i*g_.step_;
    Real i1mu = 1/(1-u);
    return i1mu*i1mu;
  }
  Complex HellTableP::getNPoles_(GammaHellTable g, Complex nt) const {
    Complex pp = 0;
    for (Poles p = g.poles_.begin(); p != g.poles_.end(); ++p)
      pp += (*p).residue/(nt - (*p).pole);
    return pp;
  }

  
} // namespace hell

